__doc__ = "Exam Questions Classes"
__version__ = "0.0.1"

from OFS.Folder import Folder
from Globals import PersistentMapping,HTMLFile
import Products
from whrandom import randint

class ExamQuestion( Folder  ):
  """ExamQuestion Base Class object"""

  meta_type = "Exam Question"
  icon="misc_/Exam/question"
  
  isQuestion=1

  default_settings={ 'weight':1.0, 'init_question_data':None }

  _properties = ( {'id':'title', 'type':'string', 'mode':'wd'}, )
    
  __ac_permissions__=((
    ('Manage Tutors Information',
     ('manage_submissions', 'manage_delSubmission',
      'manage_changeSubmission','manage_clearAll',),),
    ('Access contents information',
     ('getSetting','mark','getSubmission','getSubmissions',
      'getSubmissionRecord', 'getData',
      'submittedIds','accessedIds','weighting','question',
      'feedback','distance_metric','statistics')
     ),)
                      +Folder.__ac_permissions__
                      )
  
  manage_options=Folder.manage_options+(
    {'label':'Submissions','action':'manage_submissions',},
    )

  index_html=None

  def __init__(self, id):
    self.id = id
    self._submissions=PersistentMapping()

  def manage_delSubmission(self,userids,REQUEST=None):
    """delete submission for given user, or delete all if no user given"""
    for userid in userids:
      if self._submissions.has_key(userid):
        del self._submissions[userid]
    if REQUEST is not None:
      return self.manage_submissions(self, REQUEST)

  def manage_clearAll(self,REQUEST=None):
    """Clear all submissions"""
    self._submissions.clear()
    if REQUEST is not None:
      return self.manage_submissions(self, REQUEST)
    
  def manage_changeSubmission(self,userid,REQUEST,manage_main=None):
    """Add or chage a submission for a given id using
    values from REQUEST.form - default just takes form element
    with same id as question id"""
    self._initSubmission(userid)
    self._submissions[userid]=(
      (REQUEST.form[self.id],) + self._submissions[userid][1:]
      )
    if manage_main is None:
      return self.manage_submissions(self, REQUEST)

  def _initSubmission(self,userid):
    if not self._submissions.has_key(userid):
      self._submissions[userid]=(
        None,self.getSetting('init_question_data',userid))

  def getData(self,userid):
    self._initSubmission(userid)
    return self._submissions[userid][1]
        
  manage_submissions=HTMLFile("examQuestionSubmissions", globals())

  manage_editSubmissionForm=HTMLFile("examEditSubmissionForm",globals())
    
  def getSetting(self,name,userid=None):
    """Get a setting for the question and specified user"""
    try:
      o= getattr(self,name)
      if callable(o):
        try:
          return o(self,userid=userid)
        except StandardError,e:
          raise StandardError("Error in method "+name+" in question "+self.id+": "+str(e))
      else:
        return o
    except:
      if self.default_settings.has_key(name):
        return self.default_settings[name]
      raise StandardError("Required parameter "+name+" not defined in question "+self.id)

  def getSubmission(self,userid):
    """Retrieve copy of submission record for given user"""
    if self._submissions.has_key(userid):
      return self._submissions[userid][0]
    return None

  def getSubmissionRecord(self,userid):
    """Retrieve copy of submission record for given user"""
    if self._submissions.has_key(userid):
      return self._submissions[userid]
    return None

  def getSubmissions(self,userids=None):
    """return array of submission record copies for given userids"""
    if userids is None: userids=self.submittedIds()
    submissions=[]
    for userid in userids: submissions.append(self.getSubmission(userid))
    return submissions

  def submittedIds(self):
    """list of all who have submitted"""
    userids=[]
    for userid in self.accessedIds():
      if self.getSubmission(userid) is not None:
        userids.append(userid)
    return userids

  def accessedIds(self):
    """list of all who have accessed - by default
    returns submittedIds as accesses are not recorded"""
    return self._submissions.keys()

  def weighting(self,userid=None):
    return self.getSetting('weight',userid)

  # abstract methods to make a question, generate its form
  # and its feedback html, distance metric between two submissions
  # and the statistics presentation of the submissions
  def mark(self,userid): return self.getSetting('M',userid)

  def question(self,userid):
    """html representation of the question form elements for given user"""
    self._initSubmission(userid)
    return self.getSetting('Q',userid)

  def feedback(self,userid): return self.getSetting('F',userid)

  def distance_metric(self,id1,id2):
    """0 of both have submitted the same answer, else 1"""
    sub1=self.getSubmission(id1)
    sub2=self.getSubmission(id2)
    if (sub1 is not None and sub2 is not None and sub1==sub2): return 0.0
    return 1.0
    
  def statistics(self,userid,userids):
    if hasattr(self,'S'):
      return getattr(self,'S')(userid=userid,userids=userids)
    return "-"
  
class ExamMCQ(ExamQuestion):
  """Exam Multiple Choice Question Class"""

  meta_type = "Exam MCQ"
  icon="misc_/Exam/mcq"
  
  default_settings= {
    'randomise_answers':1,
    'weight':1.0,
    'init_question_data':None
    }
  
  def __init__(self, id):
    ExamQuestion.__init__(self,id)
    #By default add in Q,A,E,C properties initially
    self.manage_addProperty('Q','','text')
    self.manage_addProperty('A','','lines')
    self.manage_addProperty('C',0,'int')
    self.manage_addProperty('E','','text')

  def _choiceorder(self,a):
    """Returns a (possibly randomised) array of indices for generating
    the order of answers"""
    list=range(0,len(a));
    if self.getSetting('randomise_answers'):
      newlist=[]
      while len(list)>0: 
        n=randint(0,len(list)-1)
	newlist.append(list[n])
	del list[n]
      return newlist
    else:
      return list

  def weighting(self,userid=None):
    return self.getSetting('weight',userid)*(
      len(self.getSetting('A'))/5.0)

  def mark(self,userid):
    if self._isCorrect(userid): return 1.0
    if self.getSubmission(userid)<>0: return 0.0
    return 1.0/len(self.getSetting('A',userid))

  def _answer_html(self,text,value,checked=None,disabled=None):
    """generate the entry for an answer"""
    html='<input type="radio" name="'+self.id+':int" value="'+str(value)+'"'
    if disabled:
      html=html+' disabled="1"'
    if checked:
      html=html+' checked="1"'
    return html+' > '+str(text)+'\n'

  def _isCorrect(self,userid):
    return self.getSubmission(userid)==self.getSetting('C',userid)

  def question(self,userid,disabled=0):
    """html representation of the question form for given user"""
    self._initSubmission(userid)
    html=self.getSetting('Q',userid)+"\n<p>"
    c=self.getSubmission(userid) or 0
    a=self.getSetting('A',userid)
    for i in self._choiceorder(a):
      html=html+self._answer_html(a[i],i+1,c==i+1,disabled)+'<br />'
    html=html+self._answer_html('Don''t know',0,c==0,disabled)+'</p>\n'
    return html

  def feedback(self,userid):
    """html representation of the feedback for given user"""
    html=self.question(userid,1)+'<p><strong>'
    if self._isCorrect(userid):
      html=html+'Correct'
    else:
      html=html+'Wrong'
    html=html+(': Correct answer is "'+
            str(self.getSetting('A',userid)[self.getSetting('C',userid)-1])
            +'"'+'</strong></p>\n'+self.getSetting('E',userid) +'\n')
    return html

  def statistics(self,userid,userids):
    """present statistics for this question - distribution of submissions"""
    submissions=self.getSubmissions(userids)
    html=(self.getSetting('Q',userid)+
          "\n<table border='1' cellspacing='0' cellpadding='0'>\n")
    a=self.getSetting('A',userid)
    c=self.getSetting('C',userid)
    html=html+('<tr><td width="50">'+str(len(filter( lambda u: u==0  ,submissions)))
               +'</td><td>Not Known</td></tr>')
    for i in range(len(a)):
      count=len(filter( lambda u,v=i+1: u==v  ,submissions))
      html=html+'<tr><td>'+str(count)+'</td><td>'
      if i+1==c:
        html=html+'<strong>'+str(a[i])+'</strong>'
      else:
        html=html+str(a[i])
      html=html+'</td></tr>\n'
    return html+'</table>\n'+self.getSetting('E',userid) +'\n'

class ExamNumericQ(ExamQuestion):
  """Exam Numeric Question Class"""

  meta_type = "Exam Numeric Question"
  icon="misc_/Exam/numeric"
  
  default_settings= {
    'randomise_answers':1,
    'weight':1.0,
    'tol':0.01,
    'init_question_data':None,
    'units':'',
    'format':'%.3g'
    }
  
  def __init__(self, id):
    ExamQuestion.__init__(self,id)
    #By default add in Q,E,C properties initially
    self.manage_addProperty('Q','','text')
    self.manage_addProperty('C','0.0','float')
    self.manage_addProperty('E','','text')

  def question(self,userid,disabled=0):
    """html representation of the question form for given user"""
    self._initSubmission(userid)
    html="<p>"+self.getSetting('Q',userid)+"<br><input name=\""+self.id+"\""
    if disabled: html=html+" disabled=\"1\""
    s=self.getSubmission(userid)
    if s: html=html+' value =\"%s\"' % s 
    html=html+">" +self.getSetting('units',userid)+"</p>"
    return html

  def _isCorrect(self,userid):
     sub=self.getSubmission(userid)
     if not sub: return
     c=float(self.getSetting('C',userid))
     v=float(sub)
     return (abs((v-c)/c) <=
             float(self.getSetting('tol')))

  def feedback(self,userid):
    """html representation of the feedback for given user"""
    html=self.question(userid,1)+'<p><strong>'
    if self._isCorrect(userid):
      html=html+'Correct'
    else:
      html=html+'Wrong'
    html=html+': Correct answer is '
    f=self.getSetting('format',userid)%float(self.getSetting('C',userid))
    html=html+f+' '+self.getSetting('units',userid)
    html=html+'</strong></p>\n'+self.getSetting('E',userid) +'\n'
    return html

  def mark(self,userid):
    if self._isCorrect(userid): return 1.0
    return 0.0

  def statistics(self,userid,userids=None):
    """present statistics for this question - fraction correct answers"""
    if userids is None: userids=self.submittedIds()
    c=0.0
    if len(userids)>0:
      for userid in userids:
        if self._isCorrect(userid): c=c+1.0
      c=100*c/len(userids)
    return self.getSetting('Q',userid)+('<br>%.1d' % c)+"% correct"

  def distance_metric(self,id1,id2):
    """0 of answers outside tolerence, up to 1 if identical"""
    sub1=self.getSubmission(id1)
    sub2=self.getSubmission(id2)
    if (sub1 is None or sub2 is None): return 1.0
    sub1=float(sub1)
    sub2=float(sub2)
    d=(abs(sub2-sub1)*2/(sub2+sub1))/float(self.getSetting('tol'))
    if d>1.0: d=1.0
    return d
